#include "stdafx.h"
#include "myEigen.h"
#include "common.h"


namespace cmmn {

	using namespace System;

	using namespace myLib;

	double ** x_raw;

	myMat x;

	int n, p, q;

	double * mean, * sd;


	double sqr( double v ){
			return v * v;
	}

	void swap( double & a, double & b){
			double c = a;
			a = b;
			b = c;
			return;
	}

	void VariMax::RotateByPair( int i, int j ){
		double * u, * v;
		u = new double[p];
		v = new double[p];
		for (int k = 0; k < p; k++){
				u[k] = sqr(F[k][i]) - sqr(F[k][j]);
				v[k] = 2.0 * F[k][i] * F[k][j];
		}
		double A = 0.0, B = 0.0, C = 0.0, D = 0.0;
		for (int k = 0; k < p; k++){
				A += u[k];
				B += v[k];
				C += sqr(u[k]) - sqr(v[k]);
				D += 2.0 * u[k] * v[k];
		}
		double num = D - (2.0 * A * B / p);
		double dom = C - ((sqr(A) - sqr(B)) / p);
		double theta4 = Math::Atan(num / dom);
		if (num > 0.0){
			if (theta4 < 0.0) theta4 += Math::PI;
		}else{
			if (theta4 > 0.0) theta4 += Math::PI;
		}
		double theta = theta4 / 4.0;
		for (int k = 0; k < p; k++){
			double X = F[k][i] * Math::Cos(theta) + F[k][j] * Math::Sin(theta);
			double Y = -F[k][i] * Math::Sin(theta) + F[k][j] * Math::Cos(theta);
			F[k][i] = X;
			F[k][j] = Y;
		}
	}

	void VariMax::Rotate(){
		if (q >= 2){
			for (int i = 0; i < p; i++){
				double v = 0.0;
				for (int j = 0; j < q; j++)
					v += sqr(F[i][j]);
				h[i] = Math::Sqrt(v);
				for (int j = 0; j < q; j++)
					F[i][j] /= h[i];
			}
			double vdiff;
			do {
				double ** G;
				G = new double*[p];
				for (int i = 0; i < p; i++){
					G[i] = new double[q];
					for (int j = 0; j < q; j++)
						G[i][j] = F[i][j];
				}
				for (int i = 0; i < q-1; i++)
					for (int j = i+1; j < q; j++)
						RotateByPair(i, j);
				vdiff = 0.0;
				for (int i = 0; i < p; i++)
					for (int j = 0; j < q; j++)
						vdiff += Math::Abs(G[i][j] - F[i][j]);
			}while(vdiff > 1.0e-15);
			for (int i = 0; i < p; i++)
				for (int j = 0; j < q; j++){
					F[i][j] *= h[i];
					Ftemp[i][j] = F[i][j];
				}

			for (int i = 0; i < p; i++)
				h[i] = sqr(h[i]);

			//
			//		^
			//
			for (int j = 0; j < q; j++){
				double v = 0.0;
				for (int i = 0; i < p; i++)
					v += sqr(F[i][j]);
				Contri[j] = v;
			}
			for (int j = 0; j < q-1; j++){
				int k = j;
				for (int j1 = j+1; j1 < q; j1++)
					if (Contri[k] < Contri[j1])
						k = j1;
				if (j < k){
					swap( Contri[j], Contri[k] );
					for (int j1 = 0; j1 < p; j1++)
						swap( F[j1][j], F[j1][k] );
				}
			}
		}
		return;
	}


	bool VariMax::LhLargerRh( int a, int b ){
			bool ck = false;
			int cka = 0;
			for (int i = 1; i < q; i++)
				if (Math::Abs(F[a][i]) > Math::Abs(F[a][cka])) cka = i;
			int ckb = 0;
			for (int i = 1; i < q; i++)
				if (Math::Abs(F[b][i]) > Math::Abs(F[b][ckb])) ckb = i;
			if (cka < ckb) {
				ck = true;
			}else{
				if (cka == ckb){
					if (Math::Abs(F[a][cka]) > Math::Abs(F[b][ckb])) 
						ck = true;
				}
			}

			return ck;
	}


	void SwapString( String ^ %a, String ^ %b ){
			String ^ t = a;
			a = b;
			b = t;
	}

	void VariMax::OrderF( array<String ^> ^ VLabel ){
		for (int i = 0; i < p-1; i++){
			int k = i;
			for (int j = i+1; j < p; j++)
				if (LhLargerRh(j, k)) k = j;
			if (i < k){
				SwapString( VLabel[i], VLabel[k] );
				for (int j1 = 0; j1 < q; j1++)
					swap( F[i][j1], F[k][j1] );
				swap( h[i], h[k] );
			}
		}
	}


	//
	//  Łuq͖@vQ ,1979, 9.5
	//	
	void ProMax::ObliqueProcrustes(){

		myMat APA(q, q), invAPA(q, q);
		for (int i = 0; i < q; i++)
			for (int j = i; j < q; j++){
				double v = 0.0;
				for (int k = 0; k < p; k++)
					v += A[k][i] * A[k][j];
				APA[i][j] = v;
				if (i != j ) APA[j][i] = v;
			}

		calcInvMat( APA, invAPA, q );

		myMat APAiAP(q, p);
		for (int i = 0; i < q; i++)
			for (int j = 0; j < p; j++){
				double v = 0.0;
				for (int k = 0; k < q; k++)
					v += invAPA[i][k] * A[j][k];
				APAiAP[i][j] = v;
			}

		myMat Q(q, q);
		for (int i = 0; i < q; i++)
			for (int j = 0; j < q; j++){
				double v = 0.0;
				for (int k = 0; k < p; k++)
					v += APAiAP[i][k] * C[k][j];
				Q[i][j] = v;
			}

		myMat QPQ(q, q);
		for (int i = 0; i < q; i++)
			for (int j = i; j < q; j++){
				double v = 0.0;
				for (int k = 0; k < q; k++)
					v += Q[k][i] * Q[k][j];
				QPQ[i][j] = v;
				if (i != j) QPQ[j][i] = v;
			}

		myMat invQPQ(q, q);
		calcInvMat( QPQ, invQPQ, q );

		double * D = new double[q];
		for (int i = 0; i < q; i++)
			D[i] = Math::Sqrt(invQPQ[i][i]);

		for (int i = 0; i < q; i++)
			for (int j = 0; j < q; j++)
				Lambda[i][j] = Q[i][j] * D[j];

		for (int i = 0; i < p; i++)
			for (int j = 0; j < q; j++){
				double v = 0.0;
				for (int k = 0; k < q; k++)
					v += A[i][k] * Lambda[k][j];
				B[i][j] = v;
			}
	}

	double ProMax::PowerR( double a, double r ){
		double v;
		if (a > 0.0) {
			v = Math::Pow(a, r);
		}else{
			if (a < 0.0){
				v = -Math::Pow(Math::Abs(a), r);
			}else{
				v = 0.0;
			}
		}

		return v;
	}

	bool ProMax::LhLargerRh( int a, int b ){
			bool ck = false;
			int cka = 0;
			for (int i = 1; i < q; i++)
				if (Math::Abs(B[a][i]) > Math::Abs(B[a][cka])) cka = i;

			int ckb = 0;
			for (int i = 1; i < q; i++)
				if (Math::Abs(B[b][i]) > Math::Abs(B[b][ckb])) ckb = i;

			if (cka < ckb){
					ck = true;
			}else{
				if (cka == ckb){
					if (Math::Abs(B[a][cka]) > Math::Abs(B[b][ckb])) ck = true;
				}
			}

			return ck;
	}

	void ProMax::Rotate(){
		if (q <= 1){
			for (int i = 0; i < p; i++)
				B[i][0] = A[i][0];
			Phi[0][0] = 1.0;
			return;
		}

		double ** U;
		U = new double*[p];
		for (int i = 0; i < p; i++)
			U[i] = new double[q];
		for (int i = 0; i < p; i++){
			double v = 0.0;
			for (int j = 0; j < q; j++)
				v += sqr(A[i][j]);
			v = Math::Sqrt(v);
			for (int j = 0; j < q; j++)
				U[i][j] = A[i][j] / v;
		}
		for (int i = 0; i < p; i++)
			for (int j = 0; j < q; j++)
				C[i][j] = PowerR(U[i][j], r);

		ObliqueProcrustes();

		myMat LmdPLmd(q, q);
		for (int i = 0; i < q; i++)
			for (int j = i; j < q; j++){
				double v = 0.0;
				for (int k = 0; k < q; k++)
					v += Lambda[k][i] * Lambda[k][j];
				LmdPLmd[i][j] = v;
				if (i != j) LmdPLmd[j][i] = v;
			}

		myMat vPhi(q, q);
		calcInvMat( LmdPLmd, vPhi, q );
		for (int i = 0; i < q; i++)
			for (int j = 0; j < q; j++)
				Phi[i][j] = vPhi[i][j];

		double * f_w;
		f_w = new double[q];
		for (int j = 0; j < q; j++){
			double v = 0.0;
			for (int i = 0; i < p; i++)
				v += sqr(B[i][j]);
			f_w[j] = v;
		}
		for (int j = 0; j < q-1; j++){
			int k = j;
			for (int j1 = j+1; j1 < q; j1++)
				if (f_w[k] < f_w[j1]) k = j1;
			if (j < k){
				swap(f_w[j], f_w[k]);
				for (int i = 0; i < p; i++){
					swap(B[i][j], B[i][k]);
				}
				for (int j1 = 0; j1 < q; j1++)
					swap(Phi[j1][j], Phi[j1][k]);
				for (int j1 = 0; j1 < q; j1++)
					swap(Phi[j][j1], Phi[k][j1]);
			}
		}

		S = new double*[p];
		for (int i = 0; i < p; i++)
			S[i] = new double[q];
		for (int i = 0; i < p; i++)
			for (int j = 0; j < q; j++){
				double v = 0.0;
				for (int k = 0; k < q; k++)
					v += B[i][k] * Phi[k][j];
				S[i][j] = v;
			}
	}

	void ProMax::OrderB( array<String^> ^ VLabel ){
		for (int i = 0; i < p-1; i++){
			int k = i;
			for (int j = i+1; j < p; j++)
				if (LhLargerRh(j, k)) k = j;
			if (i < k){
				SwapString( VLabel[i], VLabel[k] );
				for (int j1 = 0; j1 < q; j1++){
					swap(B[i][j1], B[k][j1]);
					swap(S[i][j1], S[k][j1]);
				}
			}
		}
	}


}
